@@ -22,12 +22,19 @@ APPLIANCES = "/core/appliances"
 DIRECT_OBJECTS = "/core/direct_objects"
 DOMAIN_OBJECTS = "/core/domain_objects"
 LOCATIONS = "/core/locations"
+NOTIFICATIONS = "/core/notifications"
 RULES = "/core/rules"
+SYSTEM = "/system"
+STATUS = "/system/status.xml"
 
-DEFAULT_TIMEOUT = 20
+DEFAULT_TIMEOUT = 30
+DEFAULT_USERNAME = "smile"
+DEFAULT_PORT = 80
 
 _LOGGER = logging.getLogger(__name__)
 
+SWITCH_GROUP_TYPES = ["switching", "report"]
+
 HOME_MEASUREMENTS = {
     "electricity_consumed": "power",
     "electricity_produced": "power",
@@ -77,18 +84,42 @@ DEVICE_MEASUREMENTS = {
 }
 
 SMILES = {
-    "smile_open_therm_v30": {"type": "thermostat", "friendly_name": "Adam",},
-    "smile_open_therm_v23": {"type": "thermostat", "friendly_name": "Adam",},
-    "smile_thermo_v40": {"type": "thermostat", "friendly_name": "Anna",},
-    "smile_thermo_v31": {"type": "thermostat", "friendly_name": "Anna",},
-    "smile_thermo_v18": {
+    "smile_open_therm_v3": {
+        "type": "thermostat",
+        "friendly_name": "Adam",
+    },
+    "smile_open_therm_v2": {
+        "type": "thermostat",
+        "friendly_name": "Adam",
+    },
+    "smile_thermo_v4": {
+        "type": "thermostat",
+        "friendly_name": "Anna",
+    },
+    "smile_thermo_v3": {
+        "type": "thermostat",
+        "friendly_name": "Anna",
+    },
+    "smile_thermo_v1": {
         "type": "thermostat",
         "friendly_name": "Anna",
         "legacy": True,
     },
-    "smile_v40": {"type": "power", "friendly_name": "P1",},
-    "smile_v33": {"type": "power", "friendly_name": "P1",},
-    "smile_v25": {"type": "power", "friendly_name": "P1", "legacy": True,},
+    "smile_v4": {
+        "type": "power",
+        "friendly_name": "P1",
+    },
+    "smile_v3": {
+        "type": "power",
+        "friendly_name": "P1",
+    },
+    "smile_v2": {
+        "type": "power",
+        "friendly_name": "P1",
+        "legacy": True,
+    },
+    "stretch_v3": {"type": "stretch", "friendly_name": "Stretch", "legacy": True},
+    "stretch_v2": {"type": "stretch", "friendly_name": "Stretch", "legacy": True},
 }
 
 
@@ -101,8 +132,8 @@ class Smile:
         self,
         host,
         password,
-        username="smile",
-        port=80,
+        username=DEFAULT_USERNAME,
+        port=DEFAULT_PORT,
         timeout=DEFAULT_TIMEOUT,
         websession: aiohttp.ClientSession = None,
     ):
@@ -121,15 +152,15 @@ class Smile:
             self.websession = websession
 
         self._auth = aiohttp.BasicAuth(username, password=password)
+        # Work-around for Stretchv2-aiohttp-deflate-error, can be removed for aiohttp v3.7
+        self._headers = {"Accept-Encoding": "gzip"}
 
         self._timeout = timeout
         self._endpoint = f"http://{host}:{str(port)}"
         self._appliances = None
-        self._direct_objects = None
         self._domain_objects = None
         self._home_location = None
         self._locations = None
-        self._smile_subtype = None
         self._smile_legacy = False
         self._thermo_master_id = None
 
@@ -142,25 +173,21 @@ class Smile:
         self.smile_version = ()
         self.notifications = {}
 
-    async def connect(self, retry=2):
+    async def connect(self):
         """Connect to Plugwise device."""
-        # pylint: disable=too-many-return-statements
-        url = f"{self._endpoint}{DOMAIN_OBJECTS}"
-        try:
-            with async_timeout.timeout(self._timeout):
-                resp = await self.websession.get(url, auth=self._auth)
-            if resp.status == 401:
-                raise self.InvalidAuthentication
-        except (asyncio.TimeoutError, aiohttp.ClientError):
-            if retry < 1:
-                _LOGGER.error("Error connecting to Plugwise", exc_info=True)
-                raise self.ConnectionFailedError
-            return await self.connect(retry - 1)
+        # pylint: disable=too-many-return-statements,raise-missing-from
+        names = []
 
-        result = await resp.text()
+        result = await self.request(DOMAIN_OBJECTS)
+        dsmrmain = result.find(".//module/protocols/dsmrmain")
+        network = result.find(".//module/protocols/network_router/network")
 
-        if "<vendor_name>Plugwise</vendor_name>" not in result:
-            if "<dsmrmain id" not in result:
+        vendor_names = result.findall(".//module/vendor_name")
+        for name in vendor_names:
+            names.append(name.text)
+
+        if "Plugwise" not in names:
+            if dsmrmain is None:
                 _LOGGER.error(
                     "Connected but expected text not returned, \
                               we got %s",
@@ -170,8 +197,7 @@ class Smile:
 
         # TODO creat this as another function NOT part of connect!
         # just using request to parse the data
-        do_xml = etree.XML(self.escape_illegal_xml_characters(result).encode())
-        gateway = do_xml.find(".//gateway")
+        gateway = result.find(".//gateway")
 
         if gateway is not None:
             if gateway.find("hostname") is not None:
@@ -180,35 +206,46 @@ class Smile:
             # Assume legacy
             self._smile_legacy = True
             # Try if it is an Anna, assuming appliance thermostat
-            anna = do_xml.find('.//appliance[type="thermostat"]')
+            anna = result.find('.//appliance[type="thermostat"]')
             # Fake insert version assuming Anna
             # couldn't find another way to identify as legacy Anna
             smile_version = "1.8.0"
             smile_model = "smile_thermo"
             if anna is None:
-                # P1 legacy
-                if "<dsmrmain id" in result:
-                    # Fake insert version assuming P1
-                    # yes we could get this from system_status
-                    smile_version = "2.5.9"
-                    smile_model = "smile"
-                    # for legacy P1 use the dsmrmain id as gateway_id
-                    dsmrmain = do_xml.find(".//dsmrmain")
-                    self.gateway_id = dsmrmain.attrib["id"]
+                # P1 legacy:
+                if dsmrmain is not None:
+                    try:
+                        status = await self.request(STATUS)
+                        smile_version = status.find(".//system/version").text
+                        smile_model = status.find(".//system/product").text
+                        self.smile_hostname = status.find(".//network/hostname").text
+                    except self.InvalidXMLError:
+                        raise self.ConnectionFailedError
+
+                # Stretch:
+                elif network is not None:
+                    try:
+                        system = await self.request(SYSTEM)
+                        smile_version = system.find(".//gateway/firmware").text
+                        smile_model = system.find(".//gateway/product").text
+                        self.smile_hostname = system.find(".//gateway/hostname").text
+                        self.gateway_id = network.attrib["id"]
+                    except self.InvalidXMLError:
+                        raise self.ConnectionFailedError
                 else:
                     _LOGGER.error("Connected but no gateway device information found")
                     raise self.ConnectionFailedError
 
         if not self._smile_legacy:
-            smile_model = do_xml.find(".//gateway/vendor_model").text
-            smile_version = do_xml.find(".//gateway/firmware_version").text
+            smile_model = result.find(".//gateway/vendor_model").text
+            smile_version = result.find(".//gateway/firmware_version").text
 
         if smile_model is None or smile_version is None:
             _LOGGER.error("Unable to find model or version information")
             raise self.UnsupportedDeviceError
 
         ver = semver.parse(smile_version)
-        target_smile = f"{smile_model}_v{ver['major']}{ver['minor']}"
+        target_smile = f"{smile_model}_v{ver['major']}"
 
         _LOGGER.debug("Plugwise identified as %s", target_smile)
 
@@ -243,10 +280,15 @@ class Smile:
         await self.websession.close()
 
     async def request(
-        self, command, retry=3, method="get", data=None, headers=None,
+        self,
+        command,
+        retry=3,
+        method="get",
+        data=None,
+        headers=None,
     ):
         """Request data."""
-        # pylint: disable=too-many-return-statements
+        # pylint: disable=too-many-return-statements,raise-missing-from
 
         url = f"{self._endpoint}{command}"
 
@@ -256,11 +298,18 @@ class Smile:
         try:
             with async_timeout.timeout(self._timeout):
                 if method == "get":
-                    resp = await self.websession.get(url, auth=self._auth)
+                    # Work-around, see above, can be removed for aiohttp v3.7:
+                    resp = await self.websession.get(
+                        url, auth=self._auth, headers=self._headers
+                    )
                 if method == "put":
                     resp = await self.websession.put(
                         url, data=data, headers=headers, auth=self._auth
                     )
+                if method == "delete":
+                    resp = await self.websession.delete(url, auth=self._auth)
+            if resp.status == 401:
+                raise self.InvalidAuthentication
 
         except asyncio.TimeoutError:
             if retry < 1:
@@ -268,22 +317,14 @@ class Smile:
                 raise self.DeviceTimeoutError
             return await self.request(command, retry - 1)
 
-        except aiohttp.ClientError:
-            _LOGGER.error(
-                "Error sending command to Plugwise: %s", command, exc_info=True
-            )
-            raise self.ErrorSendingCommandError
-
-        try:
-            result = await resp.text()
-        except asyncio.TimeoutError:
-            _LOGGER.error("Timed out reading response from Smile")
-            raise self.DeviceTimeoutError
-
         # Command accepted gives empty body with status 202
         if resp.status == 202:
             return
+        # Cornercase for stretch not responsing 202
+        if method == "put" and resp.status == 200:
+            return
 
+        result = await resp.text()
         if not result or "<error>" in result:
             _LOGGER.error("Smile response empty or error in %s", result)
             raise self.ResponseError
@@ -306,15 +347,6 @@ class Smile:
         if new_data is not None:
             self._appliances = new_data
 
-    async def update_direct_objects(self):
-        """Request direct_objects data."""
-        if self._smile_legacy and self.smile_type == "power":
-            return True
-
-        new_data = await self.request(DIRECT_OBJECTS)
-        if new_data is not None:
-            self._direct_objects = new_data
-
     async def update_domain_objects(self):
         """Request domain_objects data."""
         new_data = await self.request(DOMAIN_OBJECTS)
@@ -355,14 +387,6 @@ class Smile:
             _LOGGER.error("Appliance data missing")
             raise self.XMLDataMissingError
 
-        await self.update_direct_objects()
-        # P1 legacy has no direct_objects
-        if self._direct_objects is None and (
-            self.smile_type == "power" and not self._smile_legacy
-        ):
-            _LOGGER.error("Direct_objects data missing")
-            raise self.XMLDataMissingError
-
         await self.update_domain_objects()
         if self._domain_objects is None:
             _LOGGER.error("Domain_objects data missing")
@@ -402,11 +426,13 @@ class Smile:
             # Inject home_location as dev_id for legacy so
             # get_appliance_data can use loc_id for dev_id.
             appliances[self._home_location] = {
-                "name": "Smile P1",
+                "name": "P1",
                 "types": set(["power", "home"]),
                 "class": "gateway",
                 "location": home_location,
             }
+            self.gateway_id = self._home_location
+
             return appliances
 
         # TODO: add locations with members as appliance as well
@@ -457,6 +483,7 @@ class Smile:
             if (
                 appliance.find(".//actuator_functionalities/relay_functionality")
                 is not None
+                or appliance.find(".//actuators/relay") is not None
             ):
                 appliance_types.add("plug")
             elif (
@@ -479,7 +506,7 @@ class Smile:
         home_location = None
         locations = {}
 
-        # Legacy Anna without outdoor_temp has no locations, create one containing all appliances
+        # Legacy Anna without outdoor_temp and Stretches have no locations, create one containing all appliances
         if len(self._locations) == 0 and self._smile_legacy:
             appliances = set([])
             home_location = 0
@@ -488,11 +515,21 @@ class Smile:
             for appliance in self._appliances:
                 appliances.add(appliance.attrib["id"])
 
-            locations[0] = {
-                "name": "Legacy",
-                "types": set(["temperature"]),
-                "members": appliances,
-            }
+            if self.smile_type == "thermostat":
+                locations[0] = {
+                    "name": "Legacy Anna",
+                    "types": set(["temperature"]),
+                    "members": appliances,
+                }
+            if self.smile_type == "stretch":
+                locations[0] = {
+                    "name": "Legacy Stretch",
+                    "types": set(["power"]),
+                    "members": appliances,
+                }
+
+            self._home_location = home_location
+
             return locations, home_location
 
         for location in self._locations:
@@ -652,8 +689,51 @@ class Smile:
 
             devices[appliance] = details
 
+        group_data = self.get_group_switches()
+        if group_data is not None:
+            devices.update(group_data)
+
         return devices
 
+    def get_group_switches(self):
+        """Provide switching- or pump-groups, from DOMAIN_OBJECTS."""
+        switch_groups = {}
+        search = self._domain_objects
+
+        appliances = search.findall("./appliance")
+        groups = search.findall("./group")
+
+        for group in groups:
+            group_appl = {}
+            members = []
+            group_id = group.attrib["id"]
+            group_name = group.find("name").text
+            group_type = group.find("type").text
+            if self.smile_type == "stretch":
+                group_appliance = group.findall("appliances/appliance")
+                for dummy in group_appliance:
+                    members.append(dummy.attrib["id"])
+            else:
+                for appliance in appliances:
+                    if appliance.find("./groups/group") is not None:
+                        appl_id = appliance.attrib["id"]
+                        apl_gr_id = appliance.find("./groups/group").attrib["id"]
+                        if apl_gr_id == group_id:
+                            members.append(appl_id)
+
+            if group_type in SWITCH_GROUP_TYPES:
+                group_appl[group_id] = {
+                    "name": group_name,
+                    "types": {"switch_group"},
+                    "class": group_type,
+                    "members": members,
+                    "location": None,
+                }
+
+            switch_groups.update(group_appl)
+
+        return switch_groups
+
     def get_device_data(self, dev_id):
         """Provide device-data, based on location_id, from APPLIANCES."""
         devices = self.get_all_devices()
@@ -698,28 +778,37 @@ class Smile:
 
         # Anna specific
         if details["class"] in ["thermostat"]:
-            device_data["illuminance"] = self.get_object_value(
-                "appliance", dev_id, "illuminance"
-            )
+            illuminance = self.get_object_value("appliance", dev_id, "illuminance")
+            if illuminance is not None:
+                device_data["illuminance"] = illuminance
 
         # Generic
         if details["class"] == "gateway" or dev_id == self.gateway_id:
-
-            # Try to get P1 data
-            power_data = self.get_direct_objects_from_location(details["location"])
-            if power_data is not None:
-                device_data.update(power_data)
-
-            heater_data = self.get_appliance_data(self.heater_id)
-            outdoor_temperature = heater_data.get("outdoor_temperature")
-            if outdoor_temperature is None:
+            # Anna: outdoor_temperature only present in domain_objects
+            if "outdoor_temperature" not in device_data:
                 outdoor_temperature = self.get_object_value(
                     "location", self._home_location, "outdoor_temperature"
                 )
-
-            if outdoor_temperature:
+            if outdoor_temperature is not None:
                 device_data["outdoor_temperature"] = outdoor_temperature
 
+            # Try to get P1 data and 2nd outdoor_temperature, when present
+            power_data = self.get_power_data_from_location(details["location"])
+            if power_data is not None:
+                device_data.update(power_data)
+
+        ## Switching Groups
+        if details["class"] in SWITCH_GROUP_TYPES:
+            counter = 0
+            for member in details["members"]:
+                appl_data = self.get_appliance_data(member)
+                if appl_data["relay"]:
+                    counter += 1
+
+            device_data["relay"] = True
+            if counter == 0:
+                device_data["relay"] = False
+
         return device_data
 
     def get_appliance_data(self, dev_id):
@@ -787,8 +876,10 @@ class Smile:
             try:
                 if float(measure) < 10:
                     measure = float(f"{round(float(measure), 2):.2f}")
-                else:
+                elif float(measure) >= 10 and float(measure) < 100:
                     measure = float(f"{round(float(measure), 1):.1f}")
+                elif float(measure) >= 100:
+                    measure = int(round(float(measure)))
             except ValueError:
                 if measure == "on":
                     measure = True
@@ -796,18 +887,12 @@ class Smile:
                     measure = False
         return measure
 
-    def get_direct_objects_from_location(self, loc_id):
-        """
-        Obtain the appliance-data from appliances without a location.
-
-        Determined from DIRECT_OBJECTS.
-        """
+    def get_power_data_from_location(self, loc_id):
+        """Obtain the power-data from domain_objects based on location."""
         direct_data = {}
-        search = self._direct_objects
+        search = self._domain_objects
         t_string = "tariff"
-
         if self._smile_legacy and self.smile_type == "power":
-            search = self._domain_objects
             t_string = "tariff_indicator"
 
         loc_logs = search.find(f'.//location[@id="{loc_id}"]/logs')
@@ -844,21 +929,24 @@ class Smile:
                         peak = "off_peak"
                     log_found = log_type.split("_")[0]
                     key_string = f"{measurement}_{peak}_{log_found}"
+                    net_string = f"net_electricity_{log_found}"
+                    val = loc_logs.find(locator).text
+                    f_val = self._format_measure(val)
                     if "gas" in measurement:
                         key_string = f"{measurement}_{log_found}"
-                    net_string = f"net_electricity_{log_found}"
-                    val = float(loc_logs.find(locator).text)
+                        f_val = float(f"{round(float(val), 3):.3f}")
 
                     # Energy differential
                     if "electricity" in measurement:
+                        f_val = float(f"{round(float(val), 1):.1f}")
                         diff = 1
                         if "produced" in measurement:
                             diff = -1
                         if net_string not in direct_data:
                             direct_data[net_string] = float()
-                        direct_data[net_string] += float(val * diff)
+                        direct_data[net_string] += float(f_val * diff)
 
-                    direct_data[key_string] = val
+                    direct_data[key_string] = f_val
 
         if direct_data != {}:
             return direct_data
@@ -874,7 +962,7 @@ class Smile:
                 "rule[active='true']/directives/when/then"
             )
             if active_rule is None or "icon" not in active_rule.keys():
-                return "none"
+                return
             return active_rule.attrib["icon"]
 
         locator = f'.//location[@id="{loc_id}"]/preset'
@@ -894,7 +982,7 @@ class Smile:
         if rule_ids is None:
             rule_ids = self.get_rule_ids_by_name("Thermostat presets", loc_id)
             if rule_ids is None:
-                return None
+                return presets
 
         for rule_id in rule_ids:
             directives = self._domain_objects.find(f'rule[@id="{rule_id}"]/directives')
@@ -949,15 +1037,14 @@ class Smile:
         if rule_ids is None:
             return available, selected, schedule_temperature
 
-        for rule_id, location_id in rule_ids.items():
+        for rule_id, dummy in rule_ids.items():
             active = False
             name = self._domain_objects.find(f'rule[@id="{rule_id}"]/name').text
-            if location_id == loc_id:
-                if (
-                    self._domain_objects.find(f'rule[@id="{rule_id}"]/active').text
-                    == "true"
-                ):
-                    active = True
+            if (
+                self._domain_objects.find(f'rule[@id="{rule_id}"]/active').text
+                == "true"
+            ):
+                active = True
             schemas[name] = active
             schedules = {}
             days = {
@@ -1031,20 +1118,19 @@ class Smile:
         tag = "zone_preset_based_on_time_and_presence_with_override"
 
         rule_ids = self.get_rule_ids_by_tag(tag, loc_id)
-        if rule_ids is not None:
-            for rule_id, location_id in rule_ids.items():
-                if location_id == loc_id:
-                    schema_name = self._domain_objects.find(
-                        f'rule[@id="{rule_id}"]/name'
-                    ).text
-                    schema_date = self._domain_objects.find(
-                        f'rule[@id="{rule_id}"]/modified_date'
-                    ).text
-                    schema_time = parse(schema_date)
-                    schemas[schema_name] = (schema_time - epoch).total_seconds()
+        if rule_ids is None:
+            return
 
-            if schemas != {}:
-                last_modified = sorted(schemas.items(), key=lambda kv: kv[1])[-1][0]
+        for rule_id, dummy in rule_ids.items():
+            schema_name = self._domain_objects.find(f'rule[@id="{rule_id}"]/name').text
+            schema_date = self._domain_objects.find(
+                f'rule[@id="{rule_id}"]/modified_date'
+            ).text
+            schema_time = parse(schema_date)
+            schemas[schema_name] = (schema_time - epoch).total_seconds()
+
+        if schemas != {}:
+            last_modified = sorted(schemas.items(), key=lambda kv: kv[1])[-1][0]
 
         return last_modified
 
@@ -1057,8 +1143,6 @@ class Smile:
             if rule.find(locator1) is not None:
                 if rule.find(locator2) is not None:
                     schema_ids[rule.attrib["id"]] = loc_id
-                else:
-                    schema_ids[rule.attrib["id"]] = None
 
         if schema_ids != {}:
             return schema_ids
@@ -1074,22 +1158,20 @@ class Smile:
         if schema_ids != {}:
             return schema_ids
 
-    def get_object_value(self, obj_type, appl_id, measurement):
-        """Obtain the illuminance value from the thermostat."""
-        search = self._direct_objects
-
-        if self._smile_legacy and self.smile_type == "power":
-            search = self._domain_objects
+    def get_object_value(self, obj_type, obj_id, measurement):
+        """Obtain the object-value from the thermostat."""
+        search = self._domain_objects
 
         locator = (
-            f'.//{obj_type}[@id="{appl_id}"]/logs/point_log'
+            f'.//{obj_type}[@id="{obj_id}"]/logs/point_log'
             f'[type="{measurement}"]/period/measurement'
         )
         if search.find(locator) is not None:
-            val = float(f"{round(float(search.find(locator).text), 1):.1f}")
-
+            val = self._format_measure(search.find(locator).text)
             return val
 
+        return None
+
     async def set_schedule_state(self, loc_id, name, state):
         """
         Set the schedule, with the given name, connected to a location.
@@ -1166,15 +1248,37 @@ class Smile:
 
         return f"{LOCATIONS};id={loc_id}/thermostat;id={thermostat_functionality_id}"
 
-    async def set_relay_state(self, appl_id, state):
+    async def set_relay_state(self, appl_id, members, state):
         """Switch the Plug off/on."""
-        locator = (
-            f'appliance[@id="{appl_id}"]/actuator_functionalities/relay_functionality'
+        actuator = "actuator_functionalities"
+        relay = "relay_functionality"
+        stretch_v2 = (
+            self.smile_type == "stretch" and self.smile_version[1]["major"] == 2
         )
+        if stretch_v2:
+            actuator = "actuators"
+            relay = "relay"
+
+        if members is not None:
+            for member in members:
+                locator = f'appliance[@id="{member}"]/{actuator}/{relay}'
+                relay_functionality_id = self._appliances.find(locator).attrib["id"]
+                uri = f"{APPLIANCES};id={member}/relay;id={relay_functionality_id}"
+                if stretch_v2:
+                    uri = f"{APPLIANCES};id={member}/relay"
+                state = str(state)
+                data = f"<{relay}><state>{state}</state></{relay}>"
+
+                await self.request(uri, method="put", data=data)
+            return True
+
+        locator = f'appliance[@id="{appl_id}"]/{actuator}/{relay}'
         relay_functionality_id = self._appliances.find(locator).attrib["id"]
         uri = f"{APPLIANCES};id={appl_id}/relay;id={relay_functionality_id}"
+        if stretch_v2:
+            uri = f"{APPLIANCES};id={appl_id}/relay"
         state = str(state)
-        data = f"<relay_functionality><state>{state}</state></relay_functionality>"
+        data = f"<{relay}><state>{state}</state></{relay}>"
 
         await self.request(uri, method="put", data=data)
         return True
@@ -1245,6 +1349,14 @@ class Smile:
         await self.request(uri, method="put", data=data)
         return True
 
+    async def delete_notification(self):
+        """Send a set request to the schema with the given name."""
+        uri = f"{NOTIFICATIONS}"
+        
+        await self.request(uri, method="delete")
+        return True
+
+
     class PlugwiseError(Exception):
         """Plugwise exceptions class."""
 
